from pymediainfo import MediaInfo
from typing import Dict, Sequence, Union

MediaAttribute = Union[str, bool, int, float]
MediaTrack = Dict[str, MediaAttribute]
MediaInfoList = Sequence[MediaTrack]


def get_media_info_list(filename: str) -> MediaInfoList:
    return [x.to_data() for x in MediaInfo.parse(filename).tracks]



import json
import os

from pathlib import Path
from typing import cast, Dict, Optional, Sequence, Union
from typing_extensions import TypedDict

from .attribute import Attribute, SingleAttribute, CompositeAttribute
from .preference import Preference, SinglePreference
from .level import Level, DEFAULT_LEVEL


LevelMap = Dict[str, Level]


class ConfigPreference(TypedDict):
    title: str
    level: str
    pattern: Optional[str]


class ConfigSingleAttribute(TypedDict):
    title: str
    preferences: Sequence[ConfigPreference]
    track_type: str
    track_attribute: str
    render: Optional[str]
    default_level: Optional[str]
    missing_value: Optional[str]


class ConfigCompositeAttribute(TypedDict):
    title: str
    attributes: Sequence[ConfigSingleAttribute]  # noqa: F821
    preferences: Sequence[ConfigPreference]
    render: Optional[str]
    default_level: Optional[str]


ConfigAttribute = Union[ConfigSingleAttribute, ConfigCompositeAttribute]


class ConfigLevel(TypedDict):
    color: str
    flag: bool


class Config(TypedDict):
    attributes: Sequence[ConfigAttribute]
    levels: Dict[str, ConfigLevel]


def get_preference(config_preference: ConfigPreference, level_map: LevelMap, render: str) -> Preference:
    return SinglePreference(
        title=render % (config_preference["title"]),
        pattern=config_preference.get("pattern", None) or config_preference["title"],
        level=level_map[config_preference["level"]],
    )


def get_single_attribute(config_attribute: ConfigSingleAttribute, level_map: LevelMap) -> Attribute:
    default_level_name = config_attribute.get("default_level", None) or ""
    default_level = level_map.get(default_level_name, DEFAULT_LEVEL)

    attribute = SingleAttribute(
        title=config_attribute["title"],
        render=config_attribute.get("render", None) or "%s",
        preferences=[],
        default_level=default_level,
        track_type=config_attribute["track_type"],
        track_attribute=config_attribute["track_attribute"],
        missing_value=config_attribute.get("missing_value", None),
    )
    attribute.preferences = [
        get_preference(x, level_map, attribute.render) for x in config_attribute.get("preferences", [])
    ]
    return attribute


def get_composite_attribute(config_attribute: ConfigCompositeAttribute, level_map: LevelMap) -> Attribute:
    default_level_name = config_attribute.get("default_level", None) or ""
    default_level = level_map.get(default_level_name, DEFAULT_LEVEL)

    attribute = CompositeAttribute(
        title=config_attribute["title"],
        render=config_attribute.get("render", None) or "%s",
        attributes=[],
        preferences=[],
        default_level=default_level,
    )
    attribute.attributes = [get_attribute(x, level_map) for x in config_attribute.get("attributes", [])]
    attribute.preferences = [
        get_preference(x, level_map, attribute.render) for x in config_attribute.get("preferences", [])
    ]
    return attribute


def get_attribute(config_attribute: ConfigAttribute, level_map: LevelMap) -> Attribute:
    return (
        get_composite_attribute(cast(ConfigCompositeAttribute, config_attribute), level_map)
        if "attributes" in config_attribute
        else get_single_attribute(cast(ConfigSingleAttribute, config_attribute), level_map)
    )


def get_default_config() -> Config:
    path = Path(__file__)
    config = path.parent.joinpath("default_config.json")
    return cast(Config, json.loads(config.read_text()))


def save_config(config: Config, path: Path) -> None:
    dir = path.parent
    if not dir.is_dir():
        os.makedirs(str(dir))
    config = get_default_config()
    path.write_text(json.dumps(config, indent=4))


def load_config(path: Path) -> Config:
    return cast(Config, json.loads(path.read_text()))


def make_attributes(config: Config = get_default_config()) -> Sequence[Attribute]:
    level_map = {}
    for name, options in config["levels"].items():
        level_map[name] = Level(color=options["color"], flag=options["flag"])

    attributes = []

    for config_attribute in config["attributes"]:
        attribute = get_attribute(config_attribute, level_map)
        attributes.append(attribute)

    return attributes


def get_attributes(path: Path) -> Sequence[Attribute]:
    if path.exists():
        config = load_config(path)
    else:
        config = get_default_config()
        save_config(config, path)

    return make_attributes(config)



import re

from abc import abstractmethod, ABC
from dataclasses import dataclass
from typing import List, Optional, Sequence

from .exceptions import MissingAttributeError
from .level import Level, DEFAULT_LEVEL
from .mediainfo import MediaInfoList
from .preference import Preference, SinglePreference


class Attribute(ABC):
    @abstractmethod
    def get_preference(self, media_info: MediaInfoList) -> Preference:
        raise NotImplementedError

    @abstractmethod
    def get_title(self) -> str:
        raise NotImplementedError

    @abstractmethod
    def get_preferences(self) -> Sequence[Preference]:
        raise NotImplementedError


@dataclass
class SingleAttribute(Attribute):
    preferences: List[Preference]
    title: str
    track_type: str
    track_attribute: str
    default_level: Level = DEFAULT_LEVEL
    render: str = "%s"
    missing_value: Optional[str] = None

    def get_value(self, media_info: MediaInfoList) -> Optional[str]:
        for track in media_info:
            if track["track_type"] == self.track_type:
                value = track.get(self.track_attribute, None)
                return None if value is None else str(value)

        return None

    def get_preference(self, media_info: MediaInfoList) -> Preference:
        value = self.get_value(media_info) or self.missing_value

        if not value:
            raise MissingAttributeError(f"Value missing for {self.track_type}:{self.track_attribute}")

        for preference in self.preferences:
            if preference.match(value):
                return preference

        preference = SinglePreference(
            level=self.default_level, title=self.render % (value), pattern=re.escape(value)
        )
        self.preferences.append(preference)
        return preference

    def get_title(self) -> str:
        return self.title

    def get_preferences(self) -> Sequence[Preference]:
        return self.preferences


@dataclass
class CompositeAttribute(Attribute):
    preferences: List[Preference]
    attributes: List[Attribute]
    title: str
    default_level: Level = DEFAULT_LEVEL
    render: str = "%s"

    def get_preference(self, media_info: MediaInfoList) -> Preference:
        all_values_missing = True
        value_list = []
        for attribute in self.attributes:
            try:
                value_list.append(attribute.get_preference(media_info).get_title())
                all_values_missing = False
            except MissingAttributeError:
                pass

        if all_values_missing:
            raise MissingAttributeError(f"All values missing for composite attribute '{self.title}''")

        value = "".join(value_list)

        for preference in self.preferences:
            if preference.match(value):
                return preference

        preference = SinglePreference(
            level=self.default_level, title=self.render % (value), pattern=re.escape(value)
        )
        self.preferences.append(preference)
        return preference

    def get_title(self) -> str:
        return self.title

    def get_preferences(self) -> Sequence[Preference]:
        return self.preferences



import json
import os

from abc import abstractmethod, ABC
from pathlib import Path
from sqlite3 import Connection
from typing import List, Optional, Sequence

from .attribute import Attribute
from .db import get_tracks
from .exceptions import MissingAttributeError
from .mediainfo import get_media_info_list, MediaInfoList
from .quality import Quality


class Video(ABC):
    @abstractmethod
    def get_path(self) -> Path:
        raise NotImplementedError

    @abstractmethod
    def get_qualities(self) -> Sequence[Quality]:
        raise NotImplementedError

    @abstractmethod
    def get_cached_media_info_list(self, connection: Connection) -> MediaInfoList:
        raise NotImplementedError

    @abstractmethod
    def is_flagged(self) -> bool:
        raise NotImplementedError

    @abstractmethod
    def analyze(self, attributes: Sequence[Attribute], connection: Connection) -> None:
        raise NotImplementedError


class SingleVideo(Video):
    path: Path
    filename: str
    qualities: List[Quality]

    def __init__(self, path: Path, qualities: Optional[List[Quality]] = None):
        self.path = path
        self.filename = str(path.absolute())
        self.qualities = qualities or []

    def get_path(self) -> Path:
        return self.path

    def get_qualities(self) -> Sequence[Quality]:
        return self.qualities

    def is_flagged(self) -> bool:
        for quality in self.qualities:
            if quality.preference.is_flagged():
                return True

        return False

    def get_media_info_list(self) -> MediaInfoList:
        return get_media_info_list(self.filename)

    def get_cached_media_info_list(self, connection: Connection) -> MediaInfoList:
        size = os.path.getsize(self.filename)
        modified = os.path.getmtime(self.filename)

        return get_tracks(connection, self.filename, size, modified, lambda: self.get_media_info_list())

    def analyze(self, attributes: Sequence[Attribute], connection: Connection) -> None:
        tracks = self.get_cached_media_info_list(connection)

        for attribute in attributes:
            try:
                self.qualities.append(
                    Quality(attribute=attribute, preference=attribute.get_preference(tracks))
                )
            except MissingAttributeError:
                continue
            except Exception as e:
                print(json.dumps(tracks, indent=4))
                raise (e)



import colored
from colored import stylize
from dataclasses import dataclass
from typing import cast


@dataclass
class Level:
    color: str = ""
    flag: bool = False

    def render(self, input: str) -> str:
        if not self.color:
            return input
        return cast(str, stylize(input, colored.fg(self.color)))


DEFAULT_LEVEL = Level()



import json
import os
import sqlite3

from pathlib import Path
from typing import Callable

from .mediainfo import MediaInfoList


def get_connection(path: Path) -> sqlite3.Connection:
    dir = path.parent
    if not dir.is_dir():
        os.makedirs(str(dir))

    connection = sqlite3.connect(str(path))
    cursor = connection.cursor()
    cursor.execute("SELECT count(name) FROM sqlite_master WHERE type='table' AND name='videos'")

    if cursor.fetchone()[0] == 0:
        cursor.execute(
            "CREATE TABLE videos (filename varchar primary key, size integer, modified real, tracks text)"
        )

    cursor.close()
    return connection


def get_tracks(
    connection: sqlite3.Connection,
    filename: str,
    size: int,
    modified: float,
    tracks_generator: Callable[[], MediaInfoList],
) -> MediaInfoList:
    cursor = connection.cursor()
    cursor.execute("SELECT * FROM videos WHERE filename=?", [filename])
    record = cursor.fetchone()

    if not record:
        tracks = tracks_generator()
        with connection:
            cursor.execute(
                "INSERT INTO videos VALUES (?,?,?,?)",
                [filename, size, modified, json.dumps(tracks)],
            )
    elif record[1] != size or record[2] != modified:
        tracks = tracks_generator()
        with connection:
            cursor.execute(
                "UPDATE videos SET size = ?, modified = ?, tracks = ? WHERE filename = ?",
                [size, modified, json.dumps(tracks), filename],
            )
    else:
        tracks = json.loads(record[3])

    cursor.close()

    return tracks
